<!---
	components.publishing.service
	
	This component provides a function getContent(), that allows content consumers to retrieve
	content from a repository.

	There are two kinds of rule, user-saved or programatic:
	
	* 	User-saved rules are created when a client calls getContent and specifies the
		"saveAsRule" attribute.  The arguments to getContent are saved as a WDDX file
		in the rules directory called rulename.wddx
		
	*	Each programatic rule is created by a programmer as rulename.cfc in the rules directory.
		Each cfc must implement a function:
		
			query rule(stParams)
			
		The programmer can use <cfproperty> tags in the rule cfc to cue users of the rule what parameters 
		are required in stParams
		
	This approach gives us the most flexibility in the kinds of rules we can create.
	
	The criteria argument can contain two types of parameter placeholder:
	
	*	$Parameter$ - substituted with value of parameter in stParams
	
	*	$Rule(parameter1=value1, parameter2=value2, ..., parameterN=valueN)$
			
--->

<cfcomponent displayname="Publishing Service"
	hint="Functions to retrieve content and manage rules for retrieving content">
	
	<!--- PUBLIC FUNCTIONS --->
	
	<cffunction name="getContent"
		access="remote"
		hint="Retrive content specified in criteria from repository in specified format">
		
		<cfargument name="source" 
			type="string" 
			default="" 
			required="true"
			hint="Format is 'dataStoreType:datasource' or 'rule[:ruleName]'">
			
		<cfargument name="criteria" 
			type="string" 
			default="" 
			required="false"
			hint="Optional.  Search criteria applicable to data store type, e.g. SQL or Verity search terms">

		<cfargument name="stParams" 
			type="structure"  
			required="false"
			hint="Optional.  Structure containing parameters to pass to rule or substitute in criteria string for $parameter name$">
			
		<cfargument name="returnFormat"
			type="string"
			default="query"
			required="false"
			hint="One of the following: query (default), arrayOfInstances, arrayOfStructures, XML, WDDX">
			
		<cfargument name="maxResults" 
			type="numeric"  
			required="false"
			hint="Optional.  Max results to return.">
			
		<cfargument name="resetCache" 
			type="boolean"  
			default="false"
			hint="If true, don't use cached results, fetch results again and update the cache.">

		<cfargument name="saveAsRule" 
			type="string" 
			default="" 
			required="false"
			hint="Optional.  Save this set of getContent arguments as a rule under this name">
		
		<!--- Get tickcount for performance reporting --->
		<cfset tickStart = getTickCount()>
		
		<!--- Validate Arguments --->
				
		<cfif listFindNoCase("query,arrayOfInstances,arrayOfStructures,XML,WDDX", arguments.returnFormat) eq 0>
		
			<cfthrow message="components.publishing.service.getContent: returnFormat '#arguments.returnFormat#' should be one of the following: query (default), arrayOfInstances, arrayOfStructures, XML, WDDX">
		
		</cfif>
		
		<!--- Save arguments as a rule if requested --->
		
		<cfif arguments.saveAsRule neq "">
		
			<cfwddx action="cfml2wddx" 
				input=#arguments#
				output="wArguments">
				
			<cffile action="write" 
				output=#wArguments# 
				file=#ExpandPath("/rules/" & arguments.saveAsRule & ".wddx")#
				nameconflict="overwrite">
		
		</cfif>
		
		<!--- Check cache --->
		
		<cfset bSave = false>
		<cfset cacheKey = hashArguments(arguments)>
		
		<cflock scope="APPLICATION" type="READONLY" timeout=1>
		
			<cfset bCacheStructExists = isDefined("application.PECache")>
			
			<cfif bCacheStructExists>
	
				<cfset bCacheExists = isdefined("application.PECache." & cacheKey)>
			
				<cfif bCacheExists>
			
					<cfset result = application.PECache[cacheKey]>
				
				</cfif>
			
			</cfif>
		
		</cflock>
		
		<cfif bCacheExists>
			
			<cfif arguments.resetCache>
			
				<!--- If resetcache, clear existing cache contents --->
				
				<cflock scope="APPLICATION" type="EXCLUSIVE" timeout=1>
				
					<cfset void = structDelete(application.PECache, cacheKey)>
				
				</cflock>
				
				<cfset bSave = true>
				
			</cfif>
		
		<cfelse>
		
			<!--- Content not cached, need to save --->
			<cfset bSave = true>
		
		</cfif>
		
		<cfif bSave>
		
			<!--- Need to actually get the content, pass our arguments through to private _getContent function --->
			
			<cftry>
			
				<cfinvoke method="_getContent"
					returnVariable="result"
					argumentcollection=#arguments#/>		
			
			<cfcatch type="any">
			
				<cfset result = "components.publishing.service: ERROR - #cfcatch.message#">
				<cfset bSave = false>
			
			</cfcatch>
			</cftry>

				
			<!--- Update cache --->
			
			<cfif bSave>
			
				<cflock scope="APPLICATION" type="EXCLUSIVE" timeout=1>
				
					<cfset application.PECache[cacheKey] = result>
				
				</cflock>
			
			</cfif>
		
		</cfif>
			
		<!--- Use CFLOG to record request details --->
		<cfset requestTime = getTickCount() - tickStart>
		
		<!--- Return results --->
		<cfreturn result>
		
	</cffunction>
	
	<cffunction name="getBatchedContent"
		access="remote"
		hint="Aggregate multiple calls to getContent into a single call"
		returntype="array">
		
		<cfargument name="aParams" 
			type="array" 
			required="true"
			hint="Array of structures, each one containing a set of parameters for getContent">
		
		<cfloop from=1 to=#arrayLen(arguments.aParams)# index="i">
		
			<cfinvoke method="getContent"
				returnVariable="arguments.aParams[#i#].result"
				argumentcollection=#arguments.aParams[i]#/>
			
		</cfloop>
		
		<cfreturn arguments.aParams>
		
	</cffunction>
	
	<!--- PRIVATE FUNCTIONS --->
	
	<cffunction name="_getContent"
		access="private"
		hint="Retrive content specified in criteria from repository in specified format">
		
		<cfargument name="source" 
			type="string" 
			default="" 
			required="true"
			hint="Format is 'dataStoreType:datasource' or 'rule[:ruleName]'">
			
		<cfargument 
			name="criteria" 
			type="string" 
			default="" 
			required="false"
			hint="Optional.  Search criteria applicable to data store type, e.g. SQL or Verity search terms">

		<cfargument 
			name="stParams" 
			type="struct"  
			required="false"
			hint="Optional.  Structure containing parameters to pass to rule or substitute in criteria string for $parameter name$">
			
		<cfargument name="returnFormat"
			type="string"
			default="query"
			required="false"
			hint="One of the following: query (default), arrayOfInstances, arrayOfStructures, XML, WDDX">	
		
		<cfargument 
			name="maxResults" 
			type="numeric"  
			required="false"
			hint="Optional.  Max results to return.">
		
		<!--- Validate Arguments --->
		
		<cfscript>
		
			store=listFirst(arguments.source, ":");
			
			if(find(":", arguments.source) neq 0)
				source=listLast(arguments.source, ":");
			else
				source="";
			
		</cfscript>
		
		<!--- Is the caller invoking a specific rule using scope=rule:ruleName? --->
		
		<cfif store eq "rule" and source neq "">
		
			<!--- Find the rule in the /rules directory --->
			
			<cfdirectory action="list" 
				directory=#getCurrentTemplatePath()# 
				filter="#source#.*" 
				name="qRules">
			
			<cfif qRules.recordCount eq 0>
			
				<cfthrow message="components.publishing.service.getContent: '#source#' - no such rule">
			
			</cfif>
			
			<cfswitch expression=#listLast(qRules.name, ".")#>
			
				<cfcase value="wddx">
				
					<!--- It is a user-saved rule, call _getContent, overriding saved stParam keys with any supplied to us --->				
					
					<cffile action="read" 
						file=#ExpandPath("/rules/" & qRules.name)#
						variable="wRuleArgs">
						
					<cfwddx action="wddx2cfml"
						 input=#wRuleArgs#
						 output="stRuleArgs">
						 
					<cfscript>
			
						for (param in arguments.stParams) {
							stRuleArgs.stParams[param] = arguments.stParams[param];
						}
			
					</cfscript>
					
					<cfinvoke method="_getContent"
						returnType="query"
						returnVariable="result"
						argumentcollection=#stRuleArgs#/>
				
				</cfcase>
			
				<cfcase value="cfc">
				
					<!--- It is a programatic rule, put in a call to the rule cfc's rule method --->
					
					<cfinvoke component="components.publishing.rules.#listFirst(qRules.name, ".")#"
						method="rule"
						argumentCollection=#arguments.stParams#
						returnvariable="result"/>
					
				</cfcase>
			
				<cfdefaultcase>
				
					<cfthrow message="components.publishing.service.getContent: 'rules/#qRules.name#' is not a cfc or wddx file">
				
				</cfdefaultcase>
			
			</cfswitch>

		<cfelse>
		
			<!--- Substitute Parameters in criteria, if any --->
			
			<cfscript>
			
				for (param in arguments.stParams) {
					arguments.criteria = ReplaceNoCase(arguments.criteria, "$" & param & "$", arguments.stParams[param], "ALL");
				}
			
			</cfscript>
			
			<cfif store eq "rule">
			
				<!--- 	The criteria argument contains SQL to execute as a query of queries.  To get the source queries,
						We need to search the criteria looking for rule invocations in the form
						
							$Rule(parameter1=value1, parameter2=value2, ..., parameterN=valueN)$
						
						(Note: in this implementation, value1..valueN are strings or numbers, no nested calls to rules supported)
						For each rule that we find, we parse the parameters and call the rule (overriding the returnType to query).
						If the rule returns a string or number, we substitute the return value for the rule in the criteria.  If 
						the rule returns a query, we save the results in a variable then substitute the variable name for the rule
						in the criteria.  Once all the rules have been processed, we can pass the criteria to a cfquery tag for
						normal query of queries processing.
				--->
				
				<cfscript>
				
					nextQueryPostfix = 0;				// Incremental postfix to append to created queries
					criteriaIn = arguments.criteria;	// Criteria string to process
					criteriaOut = "";					// Processed criteria string
					lErrors = "";						// Error messages, comma delimited
					
					// Keep processing the criteria string till there's none left
					
					while (len(critetiaIn) gt 0) {
					
						// Break the criteria into chunks at the dollar signs
						chunk = listFirst(criteriaIn, "$");
						criteriaIn = listRest(criteriaIn, "$");
						
						// Does this chunk look like a rule, i.e. "name(...)"?
						stSubExpr = REFindNoCase("([_a-zA-Z0-9]+)\(([^)]*)\)", s, 1, true);
						
						if ((stSubExpr.len[1] eq len(Chunk)) and (arrayLen(stSubExpr.len) eq 3)) {
						
							// yes, attempt to parse it. Subexpressions 2 & 3 should contain
							// rule name and parameter list, respectively
							ruleName = "rule:" & mid(chunk, stSubExpr.pos[2], stSubExpr.len[2]);
							lParams = mid(chunk, stSubExpr.pos[3], stSubExpr.len[3]);
							
							// build an arguments structure out of the parameter list
							stArgs = structNew();
							
							while(len(lParams) gt 0) {
							
								keyValuePair = listFirst(lParams);
								lParams = listRest(lParams);
								
								if listLen(keyValuePair, "=") neq 2)
									lErrors = listAppend(lErrors, "'#keyValuePair#' is not a valid parameter name/value");
								else
									stArgs[trim(listFirst(keyValuePair, "="))] = trim(listLast(keyValuePair, "="));
							}
							
							// call the rule
							ruleResult = _getContent(source=ruleName stParams=stArgs returnFormat="query");
							
							if (isSimpleValue(ruleResult)) {
							
								// interpolate the result into the criteria string
								criteriaOut = criteriaOut & ruleResult;
							
							} else if (isQuery(ruleResult)) {
							
								// store the resultset in a variable, interpolate variable name into criteria string
								queryName = "qRule" & incrementValue(nextQueryPostfix);
								setVariable(queryName, ruleResult);
								criteriaOut = criteriaOut & queryName;
								
							} else {
							
								lErrors = listAppend(lErrors, "'#ruleName#' did not return a query or simple value");
							
							}
							
						} else {
							
							// no, pass it through to criteriaOut.
							criteriaOut = criteriaOut & chunk;
						
						}
					
					}
				
				</cfscript>
				
				<cfif lErrors neq "">
				
					<!--- Throw errors raised during parsing --->
					<cfthrow message="components.publishing.service.getContent: One or more errors encountered while parsing criteria - #lErrors#">
					
				<cfelse>
				
					<!--- criteriaOut should contain a valid query of queries expression --->
					<cfquery name="result" dbtype="query">#criteriaOut#</cfquery>
				
				</cfif>
				
			<cfelse>
			
				<!--- This is easy... call persistence layer to get result --->
				
				<cfinvoke component="components.persistence.basePersistentComponent"
					method="find"
					dataStoreType=#store#
					datasource=#source#
					criteria=#arguments.criteria#
					returntype="query"
					returnvariable="result"/>
			
			</cfif>
 
		</cfif>
		
		<!--- Convert to requested format for return --->
		<cfswitch expression=#arguments.returnFormat#>
		
			<cfcase value="query">
			
				<cfset results = qResults>
			
			</cfcase>
		
			<cfcase value="arrayOfInstances">
			
				<!--- Call basePersistentComponent.load() to convert --->
			
			</cfcase>
		
			<cfcase value="arrayOfStructures">
			
				<!--- Iterate through query rows, convert each row to structure --->
			
			</cfcase>
		
			<cfcase value="XML">
			
				<!--- Simple rendering of result set as an XML string --->
			
			</cfcase>
		
			<cfcase value="WDDX">
			
				<!--- return as a Web Distributed Data eXchange string --->
				
				<cfwddx action="cfml2wddx" 
					input=#qResults#
					output="results">
					
			</cfcase>
						
		</cfswitch>
		
		<!--- Return results --->
		<cfreturn results>
		
	</cffunction>
	
	<cffunction name="hashArguments"
		access="private"
		hint="Generate hash from function arguments structure"
		returntype="string">
		
		<cfargument name="stArgs"
			 type="struct"
			 required="true">
			 
		<cfscript>
		
			concatArgs = "";
			
			for (key in argument.stArgs) {
				if (isSimpleValue(argument.stArgs[key]))
					concatArgs = concatArgs & argument.stArgs;
				else if (isStruct(argument.stArgs[key]))
					concatArgs = concatArgs & hashArguments(argument.stArgs[key]);
			}
			
			<cfreturn hash(concatArgs)>
		
		</cfscript>	 
			 
	</cffunction>
	
</cfcomponent>