<!---
AUTHOR: Anthony C. McClure
DATE: 4/25/2002
PURPOSE: Database persistor component.  This component hold the logic to persist objects and types
		into a database.  This CFC should be instantiated from the basepersistentcomponent.cfc and
		called only from there through the exposed persistence API.
MODIFICATION LOG:
DATE        AUTHOR      MODIFICATION
====        =======     ============
4/25/2002	ACM			Initial Generation

This CFC contains the following methods:
	installObjectType() - creates an object type repository within the database
	removeObjectType() - removes an object type repository from within the database
	getObject() - retrieve object data for the active object.  Set properties.
	deleteObject() - remove the active object from the datastore
	updateObject() - update the active object with current properties
	createObject() - insert new object properties into the datastore.  Set properties as needed.

--->

<cfcomponent output="FALSE">
	<cffunction name="installObjectType" returnType="boolean" access="public" output="0">
		<!--- This function takes a cfc as an argument.  The cfc's properties are used to create the 
		database table for this object type through component introspection. --->
		<cfargument name="cfcObject" required="1">
		<cfargument name="dataSource" required="0" default="testDB">
		<cfset objectMetaData = getMetadata(arguments.cfcObject)>
		<cfset cfcObject.setTypeID(listLast(objectMetaData.name,"."))>
		
		<!--- Check for existing table --->
		<cftry>
		<cfquery datasource="#arguments.dataSource#" name="typeCheck">
			SELECT *
			FROM obj_#replace(cfcObject.getTypeID()," ","_","all")#
		</cfquery>
		
		<cfif typeCheck.recordCount GT -1>
			<cfreturn false/>
		</cfif>
		<cfcatch type="Any">
			<!--- Ignore on error, means no table exists --->
		</cfcatch>
		</cftry>
		
		<!--- Place all properties into a single array --->
		<cfscript>
			aAllProperties = arrayNew(1);
			for(i=1;i LTE arrayLen(objectMetaData.properties);i = i + 1) {
				arrayAppend(aAllProperties,objectMetaData.properties[i]);
			}
			for(i=1;i LTE arrayLen(objectMetaData.extends.properties);i = i + 1) {
				arrayAppend(aAllProperties,objectMetaData.extends.properties[i]);
			}
		</cfscript>

		<!--- Create a new table, using the property information --->
		<cftry>
		<cfquery name="installTable" datasource="#arguments.dataSource#">
			CREATE TABLE obj_#replace(cfcObject.getTypeID()," ","_","all")#
			(
			<cfscript>
				//loop object properties
				for(i=1;i LTE arrayLen(aAllProperties);i = i + 1) {
					//set property structure into currentProp
					currentProp = aAllProperties[i];
					
					//if no type key, it is not a property
					if (structKeyExists(currentProp,"type")) {
						//Determine if "," is needed
						if (i GT 1) 
							writeOutput(",");
						
						//create properties
						switch (currentProp.name) {
							case "objectID":
							{
								writeOutput("objectID char(35) PRIMARY KEY");
								break;
							}
							case "typeID":
							{
								writeOutput("typeID char(" & len("obj_" & cfcObject.getTypeID()) & ")");
								break;
							}
							default:
							{
								switch (currentProp.type) {
									case "string":
									{
										writeOutput(currentProp.name & " char(255)");
										break;
									}
									case "date":
									{
										writeOutput(currentProp.name & " datetime");
										break;
									}
									case "boolean":
									{
										writeOutput(currentProp.name & " byte");
										break;
									}
									case "memo":
									{
										writeOutput(currentProp.name & " memo");
										break;
									}
									case "struct":
									{
										writeOutput(currentProp.name & " memo");
										break;
									}
									case "array":
									{
										writeOutput(currentProp.name & " memo");
										break;
									}
								}
								break;
							}
						}
					}
				}
			</cfscript>
			)
		</cfquery>
		<cfcatch type="Database">
			<cfthrow message="A database error occured creating the persistence table.">
		</cfcatch>
		<cfcatch type="Any">
			<cfthrow message="An error has occured creating the persistence table.">
		</cfcatch>
		</cftry>
		
		<cfreturn true/>
	</cffunction>
	<cffunction name="createPersistenceObject" access="public" output="0" returnType="boolean">
		<!--- This function takes a cfc as an argument.  The cfc's properties are used to create the 
				object in the appropriate table. --->
		<cfargument name="cfcObject" required="1">
		<cfargument name="dataSource" required="0" default="testDB">
		<cfset objectMetaData = getMetadata(arguments.cfcObject)>
		<cfset cfcObject.setTypeID(listLast(objectMetaData.name,"."))>
		
		<!--- Place all properties into a single array --->
		<cfscript>
			aAllProperties = arrayNew(1);
			for(i=1;i LTE arrayLen(objectMetaData.properties);i = i + 1) {
				arrayAppend(aAllProperties,objectMetaData.properties[i]);
			}
			for(i=1;i LTE arrayLen(objectMetaData.extends.properties);i = i + 1) {
				arrayAppend(aAllProperties,objectMetaData.extends.properties[i]);
			}
		</cfscript>

		<!--- Create the Insert Query --->
		<cftry>
		<cfquery name="installTable" datasource="#arguments.dataSource#">
			INSERT INTO obj_#replace(cfcObject.getTypeID()," ","_","all")#
			(
			<cfscript>
				//loop object properties
				for(i=1;i LTE arrayLen(aAllProperties);i = i + 1) {
					//set property structure into currentProp
					currentProp = aAllProperties[i];
					
					//if no type key, it is not a property
					if (structKeyExists(currentProp,"type")) {
						//Determine if "," is needed
						if (i GT 1) 
							writeOutput(",");
						
						writeOutput(currentProp.name);
					}
				}
			</cfscript>
			)
			VALUES (
			<cfscript>
				//loop two for values
				for(i=1;i LTE arrayLen(aAllProperties);i = i + 1) {
					//set property structure into currentProp
					currentProp = aAllProperties[i];
					
					//if no type key, it is not a property
					if (structKeyExists(currentProp,"type")) {
						//Determine if "," is needed
						if (i GT 1) 
							writeOutput(",");
						
						//create properties
						switch (currentProp.name) {
							case "objectID":
							{
								cfcObject.setObjectID(createUUID());
								writeOutput("'" & cfcObject.getObjectID() & "'");
								break;
							}
							case "dtCreatedDate":
							{
								cfcObject.setDtCreatedDate(createODBCDate(now()));
								writeOutput(cfcObject.getDtCreatedDate());
								break;
							}
							case "dtLastUpdatedDate":
							{
								cfcObject.setDtLastUpdatedDate(createODBCDate(now()));
								writeOutput(cfcObject.getDtLastUpdatedDate());
								break;
							}
							default:
							{
								//set up the get method for the property
								evalThisProp = "cfcObject.get" & currentProp.name & "()";
								
								switch (currentProp.type) {
									case "string":
									{
										writeOutput("'" & evaluate(evalThisProp) & "'");
										break;
									}
									case "date":
									{
										writeOutput(evaluate(evalThisProp));
										break;
									}
									case "boolean":
									{
										writeOutput(evaluate(evalThisProp));
										break;
									}
									case "numeric":
									{
										writeOutput(evaluate(evalThisProp));
										break;
									}
									case "memo":
									{
										writeOutput("'" & evaluate(evalThisProp) & "'");
										break;
									}
									case "struct":
									{
										writeOutput("'" & evaluate(evalThisProp) & "'");
										break;
									}
									case "array":
									{
										writeOutput("'" & evaluate(evalThisProp) & "'");
										break;
									}
								}
								break;
							}
						}
					}
				}
			</cfscript>
			)
		</cfquery>
		<cfcatch type="Database">
			<cfthrow message="A database error occured creating the persistence table.">
		</cfcatch>
		<cfcatch type="Any">
			<cfthrow message="An error has occured creating the persistence table.">
		</cfcatch>
		</cftry>
		<cfreturn 1>
	</cffunction>
	<cffunction name="getPersistenceObject" output="1" access="public">
		<!--- This function takes a cfc as an argument.  By using the objectID and typeID, return the properties
		to the this scope from a SQL call. --->
		<cfargument name="cfcObject" required="1">
		<cfargument name="dataSource" required="0" default="testDB">
		<cfargument name="objectID" required="1">
		<cfset objectMetaData = getMetadata(arguments.cfcObject)>
		<cfset cfcObject.setTypeID(listLast(objectMetaData.name,"."))>

		<!--- Create the query --->
		<cfquery name="getObject" datasource="#arguments.dataSource#">
			SELECT *
			FROM obj_#cfcObject.getTypeID()#
			WHERE objectID = '#arguments.objectID#'
		</cfquery>

		<cfset queryProperties = getObject.columnList>
		
		<!--- Loop the query result and place into properties --->
		<cfloop query="getObject">
			<cfloop list="#queryProperties#" index="thisProperty">
				<cfset "cfcObject.#thisProperty#" = evaluate("getObject.#thisProperty#")>
			</cfloop>
		</cfloop>
		<cfreturn true>
	</cffunction>
</cfcomponent>
